"""
Base32 encoding modified from
https://github.com/mdomke/python-ulid

In turn adapted from
https://github.com/RobThree/NUlid
"""

# Note the binary order is TIMESTAMP_RANDO_PREFIX+VERSION
TIME_BIN_LEN = 5
RANDO_BIN_LEN = 8
END_RANDO_BIN = TIME_BIN_LEN + RANDO_BIN_LEN
PREFIX_BIN_LEN = 3  # includes version
BIN_LEN = TIME_BIN_LEN + RANDO_BIN_LEN + PREFIX_BIN_LEN

# But the string order is PREFIX_TIME_RANDO_VERSION
PREFIX_CHAR_LEN = 4  # excluding the version char
TIME_CHAR_LEN = 8
END_TIME_CHAR = PREFIX_CHAR_LEN + TIME_CHAR_LEN
RANDO_CHAR_LEN = 13
VERSION_CHAR_LEN = 1
CHAR_LEN = PREFIX_CHAR_LEN + TIME_CHAR_LEN + RANDO_CHAR_LEN + VERSION_CHAR_LEN

# 32-character alphabet modified from Crockford's
# Numbers first for sensible sorting, but full lower-case
# latin alphabet so any sensible prefix can be used
# Effectively a mapping from 8 bit byte -> 5 bit int -> base32 character
ENCODE = "234567abcdefghijklmnopqrstuvwxyz"

# Speedy O(1) inverse lookup
# base32 char -> ascii byte int -> base32 alphabet index
# fmt: off
DECODE = [
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
      0,   1,   2,   3,   4,   5, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255,   6,   7,   8,
      9,  10,  11,  12,  13,  14,  15,  16,  17,  18,
     19,  20,  21,  22,  23,  24,  25,  26,  27,  28,
     29,  30,  31, 255, 255, 255, 255, 255, 255, 255,
 ]
# fmt: on


def encode(binary: bytes) -> str:
    if len(binary) != BIN_LEN:
        raise ValueError(f"UPID has to be exactly {BIN_LEN} bytes long")

    time = encode_time(binary[:TIME_BIN_LEN])
    rando = encode_rando(binary[TIME_BIN_LEN:END_RANDO_BIN])
    prefix, version = encode_prefix(binary[END_RANDO_BIN:])
    return prefix + "_" + time + rando + version


def encode_prefix(binary: bytes) -> tuple[str, str]:
    if len(binary) != PREFIX_BIN_LEN:
        raise ValueError(f"Prefix value has to be exactly {PREFIX_BIN_LEN} bytes long.")
    lut = ENCODE
    return "".join(
        [
            lut[(binary[0] & 248) >> 3],
            lut[((binary[0] & 7) << 2) | ((binary[1] & 192) >> 6)],
            lut[(binary[1] & 62) >> 1],
            lut[((binary[1] & 1) << 4) | ((binary[2] & 240) >> 4)],
        ]
    ), lut[(binary[2] & 15)]  # implicitly "add" a 0 bit


def encode_time(binary: bytes) -> str:
    if len(binary) != TIME_BIN_LEN:
        raise ValueError(f"Timestamp value has to be exactly {TIME_BIN_LEN} bytes long.")
    lut = ENCODE
    return "".join(
        [
            lut[(binary[0] & 248) >> 3],
            lut[((binary[0] & 7) << 2) | ((binary[1] & 192) >> 6)],
            lut[(binary[1] & 62) >> 1],
            lut[((binary[1] & 1) << 4) | ((binary[2] & 240) >> 4)],
            lut[((binary[2] & 15) << 1) | ((binary[3] & 128) >> 7)],
            lut[(binary[3] & 124) >> 2],
            lut[((binary[3] & 3) << 3) | ((binary[4] & 224) >> 5)],
            lut[(binary[4] & 31)],
        ]
    )


def encode_rando(binary: bytes) -> str:
    if len(binary) != RANDO_BIN_LEN:
        raise ValueError(f"Randomness value has to be exactly {RANDO_BIN_LEN} bytes long.")
    lut = ENCODE
    return "".join(
        [
            lut[(binary[0] & 248) >> 3],
            lut[((binary[0] & 7) << 2) | ((binary[1] & 192) >> 6)],
            lut[(binary[1] & 62) >> 1],
            lut[((binary[1] & 1) << 4) | ((binary[2] & 240) >> 4)],
            lut[((binary[2] & 15) << 1) | ((binary[3] & 128) >> 7)],
            lut[(binary[3] & 124) >> 2],
            lut[((binary[3] & 3) << 3) | ((binary[4] & 224) >> 5)],
            lut[(binary[4] & 31)],
            lut[(binary[5] & 248) >> 3],
            lut[((binary[5] & 7) << 2) | ((binary[6] & 192) >> 6)],
            lut[(binary[6] & 62) >> 1],
            lut[((binary[6] & 1) << 4) | ((binary[7] & 240) >> 4)],
            lut[(binary[7] & 15)],  # implicitly "add" a 0 bit
        ]
    )


def decode(encoded: str) -> bytes:
    encoded = encoded.replace("_", "")
    if len(encoded) != CHAR_LEN:
        raise ValueError(f"Encoded UPID has to be exactly {CHAR_LEN} characters long.")
    if any((c not in ENCODE) for c in encoded):
        raise ValueError(f"Encoded UPID can only consist of letters in {ENCODE}.")

    prefix = decode_prefix(encoded[:PREFIX_CHAR_LEN] + encoded[-1])
    time = decode_time(encoded[PREFIX_CHAR_LEN:END_TIME_CHAR])
    rando = decode_rando(encoded[END_TIME_CHAR:-1])
    return time + rando + prefix


def decode_prefix(encoded: str) -> bytes:
    if len(encoded) != PREFIX_CHAR_LEN + VERSION_CHAR_LEN:
        raise ValueError(f"UPID prefix has to be exactly {PREFIX_CHAR_LEN} characters long.")
    lut = DECODE
    values = bytes(encoded, "ascii")
    if lut[values[-1]] > 15:
        raise ValueError(f"Prefix value {encoded} is too large and will overflow 128-bits.")
    return bytes(
        [
            ((lut[values[0]] << 3) | (lut[values[1]] >> 2)) & 0xFF,
            ((lut[values[1]] << 6) | (lut[values[2]] << 1) | (lut[values[3]] >> 4)) & 0xFF,
            ((lut[values[3]] << 4) | (lut[values[4]] & 15)) & 0xFF,
            # lose 1 bit of data
        ]
    )


def decode_time(encoded: str) -> bytes:
    if len(encoded) != TIME_CHAR_LEN:
        raise ValueError(f"UPID timestamp has to be exactly {TIME_CHAR_LEN} characters long.")
    lut = DECODE
    values: bytes = bytes(encoded, "ascii")
    return bytes(
        [
            ((lut[values[0]] << 3) | (lut[values[1]] >> 2)) & 0xFF,
            ((lut[values[1]] << 6) | (lut[values[2]] << 1) | (lut[values[3]] >> 4)) & 0xFF,
            ((lut[values[3]] << 4) | (lut[values[4]] >> 1)) & 0xFF,
            ((lut[values[4]] << 7) | (lut[values[5]] << 2) | (lut[values[6]] >> 3)) & 0xFF,
            ((lut[values[6]] << 5) | (lut[values[7]])) & 0xFF,
        ]
    )


def decode_rando(encoded: str) -> bytes:
    if len(encoded) != RANDO_CHAR_LEN:
        raise ValueError(f"UPID randomness has to be exactly {RANDO_CHAR_LEN} characters long.")
    lut = DECODE
    values = bytes(encoded, "ascii")
    if lut[values[-1]] > 15:
        raise ValueError(f"Random value {encoded} is too large and will overflow 128-bits.")
    return bytes(
        [
            ((lut[values[0]] << 3) | (lut[values[1]] >> 2)) & 0xFF,
            ((lut[values[1]] << 6) | (lut[values[2]] << 1) | (lut[values[3]] >> 4)) & 0xFF,
            ((lut[values[3]] << 4) | (lut[values[4]] >> 1)) & 0xFF,
            ((lut[values[4]] << 7) | (lut[values[5]] << 2) | (lut[values[6]] >> 3)) & 0xFF,
            ((lut[values[6]] << 5) | (lut[values[7]])) & 0xFF,
            ((lut[values[8]] << 3) | (lut[values[9]] >> 2)) & 0xFF,
            ((lut[values[9]] << 6) | (lut[values[10]] << 1) | (lut[values[11]] >> 4)) & 0xFF,
            ((lut[values[11]] << 4) | (lut[values[12]] & 15)) & 0xFF,
            # lose 1 bit of data
        ]
    )



import datetime as dt
import functools
import os
import time
import uuid
from typing import TypeVar

from upid import b32, consts

Self = TypeVar("Self", bound="UPID")

# version is 4 bits, restricted to first half of base32 alphabet
# first version is "a" purely for aesthetics
# this is not expected to change, but could allow for future
# versions with different timestamp/randomness configuration
VERSION = "a"


@functools.total_ordering
class UPID:
    """
    The `UPID` contains a 20-bit prefix, 40-bit timestamp and 68 bits of randomness.

    The prefix should only contain lower-case latin alphabet characters and be max
    four characters long.

    It is usually created using the `upid(prefix: str)` helper function:

        upid("user")  # UPID(user_3accvpp5_guht4dts56je5w)

    Note that helper methods to convert to hex/UUID/str/int are provided,
    but not always the inverse, as this may be lossy or meaningless.
    """

    b: bytes

    def __init__(self, b: bytes):
        """Not normally used directly."""
        self.b = b

    @classmethod
    def from_prefix(cls: type[Self], prefix: str) -> Self:
        """Create a new `UPID` from a `prefix`, using the current datetime."""
        milliseconds = time.time_ns() // consts.NS_PER_MS
        return cls.from_prefix_and_milliseconds(prefix, milliseconds)

    @classmethod
    def from_prefix_and_datetime(cls: type[Self], prefix: str, datetime: dt.datetime) -> Self:
        """Create a new `UPID` from a `prefix`, using the supplied `datetime`."""
        milliseconds = int(datetime.timestamp() * consts.MS_PER_SEC)
        return cls.from_prefix_and_milliseconds(prefix, milliseconds)

    @classmethod
    def from_prefix_and_milliseconds(cls: type[Self], prefix: str, milliseconds: int) -> Self:
        """
        Create a new `UPID` from a `prefix`, using the supplied `milliseconds`.

        `milliseconds` must be an int in milliseconds since the epoch.

        The timestamp is converted to 6 bytes, but we drop 1 byte, resulting
        in a time precision of about 256 milliseconds

        The prefix is padded with 'z' characters (if too short) and
        trimmed to 4 characters (if too long). Supply a prefix of exactly
        4 characters if this isn't appealing!
        """
        # we drop one byte of millisecond time information
        time_bin = int.to_bytes((milliseconds >> 8), b32.TIME_BIN_LEN + 0, "big")

        # 8 bytes/64 bits of randomness
        rando_bin = os.urandom(b32.RANDO_BIN_LEN)

        # note the padding and trimming
        prefix_clean = prefix.ljust(4, "z")[:4]
        prefix_bin = b32.decode_prefix(prefix_clean + VERSION)

        # note the binary ordering
        return cls(time_bin + rando_bin + prefix_bin)

    @classmethod
    def from_str(cls: type[Self], string: str) -> Self:
        """
        Convert the provided `str` to a `UPID`.

        Throws a `ValueError` if the string is invalid:
        - too long
        - too short
        - contains characters not in the `ENCODE` base32 alphabet
        """
        return cls(b32.decode(string))

    @property
    def prefix(self) -> str:
        """Return just the prefix as a `str`."""
        prefix, _ = b32.encode_prefix(self.b[b32.END_RANDO_BIN :])
        return prefix

    @property
    def milliseconds(self) -> int:
        """Returns a time in integer milliseconds since the epoch."""
        # must add back the deleted byte to get a sensible timestamp
        return int.from_bytes(self.b[: b32.TIME_BIN_LEN] + consts.ZERO_BYTE, "big")

    @property
    def datetime(self) -> dt.datetime:
        return dt.datetime.fromtimestamp(self.milliseconds / consts.MS_PER_SEC, dt.timezone.utc)

    @property
    def hex(self) -> str:
        return self.b.hex()

    def to_str(self) -> str:
        return b32.encode(self.b)

    def to_uuid(self) -> uuid.UUID:
        """Convert to a standard Python UUID."""
        return uuid.UUID(bytes=self.b)

    def __repr__(self) -> str:
        return f"UPID({self!s})"

    def __str__(self) -> str:
        return self.to_str()

    def __int__(self) -> int:
        return int.from_bytes(self.b, "big")

    def __bytes__(self) -> bytes:
        return self.b

    def __lt__(self, other: object) -> bool:
        if isinstance(other, UPID):
            return self.b < other.b
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, UPID):
            return self.b == other.b
        return False

    def __hash__(self) -> int:
        return hash(self.b)


def upid(prefix: str) -> UPID:
    """Generate a UPID with the provided prefix."""
    return UPID.from_prefix(prefix)
